from __future__ import absolute_import
import os.path
import sys
import logging
import vtk

from viper._error import ViperError

def _configure_logging():
    """Configure logging for Viper library.

    Since Viper is a library that might be used in an application that does not
    use the 'logging' module, we configure initially a do-nothing logger here.
    That is, nothing will be logged until the application configures at least
    one root logging handler.
    """
    global logger
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass
    logger = logging.getLogger("Viper")
    logger.addHandler(NullHandler())
    logger.setLevel(logging.DEBUG)

_configure_logging()


class NotFound(ViperError):
    pass

def resolve_path(executable):
    """ Resolve name of executable into absolute path.
    @raise NotFound: The executable was not found in path.
    @raise ValueError: An invalid filename was passed.
    """
    on_win = sys.platform.startswith("win")
    if os.path.sep in executable:
        raise ValueError("Invalid filename: %s" % executable)
    path = os.environ["PATH"].split(os.pathsep)
    if on_win:    # pragma: optional
        try: import win32api, pywintypes
        except ImportError: pass
        else:
            try: return win32api.FindExecutable(executable)[1]
            except pywintypes.error: pass

    suffixes = [""]
    if on_win:    # pragma: optional
        suffixes = os.environ["PATHEXT"].split(";")
    for d in path:
        exepath = os.path.join(d, executable)
        for suf in suffixes:
            try:
                if os.access(exepath + suf, os.X_OK):
                    return exepath
            except OSError:
                pass
    raise NotFound(executable)


def get_vtk_actors(actor_coll):
    """Iterate over a vtkActorCollection."""
    actor_coll.InitTraversal()
    while True:
        actor = actor_coll.GetNextActor()
        if actor is None:
            break
        yield actor


def iterate_vtk_array(array):
    """Iterate over a vtkAbstractArray."""
    for i in xrange(array.GetSize()):
        yield array.GetValue(i)


def get_vtk_idtype():
    """Get the numpy name of the type used for VTK L{IDs<vtkIdType>}.
    """
    return "int%d" % (vtk.vtkIdTypeArray().GetDataTypeSize() * 8,)
