#!/usr/bin/env python
from __future__ import absolute_import

__copyright__ = "(C) 2006-2008 Ola Skavhaug, Arve Knudsen and Simula Research Laboratory"
__license__   = "GNU LGPL Version 2.1."
__cite__      = """Ola Skavhaug, Viper Visualization Software, 
http://www.fenics.org/viper/"""
__version__ = "0.4.0"

r"""
Viper

A simple mesh plotter and run--time visualization module for plotting and
saving simulation data. The class C{Viper} can visualize solutions given as
numpy arrays, and meshes that provide the two methods C{cells()} and
C{coordinates()}. These methods should return numpy arrays specifying the
node-element ordering and coordinates of the nodes, respectively.

@var lutdir: Directory containing LUT files.
Copyright: %s
Licence: %s
Citation: %s
""" % (__copyright__, __license__, __cite__)

import vtk, numpy, os, math
#from vtk.util.colors import tomato
import logging
import shutil
import tempfile
import subprocess
import operator
import sys
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "dependencies"))
import configobj
import validate

import viper._signal
import viper._util
from viper._util import logger as _logger
from viper._error import ViperError
from viper._common import WidgetBase
from viper._signal import Signal


__all__ = ["ActorBase", "CellArray", "CellZeros", "ILabelable", "Mesh",
        "LookupTable", "MovieMaker", "NoFramesError", "PngWriter", "Polygon",
        "ScalarField", "Signal", "Sphere", "VectorField", "Viper", "ViperError", "VtkWindow",
        "get_default_lut", "lutdir", "set_default_lut"]


_STOCK_LUT = [[1, 0, 0, 1],
              [0, 1, 0, 1],
              [0, 0, 1, 1],
              [0, 0, 0, 1],
              [1, 1, 0, 1],
              [1, 0, 1, 1],
              [0, 1, 1, 1]
              ]


def _read_lut_file(fname, lut):
    try: lut_file = file(fname)
    except EnvironmentError:
        _logger.warning("can't open LUT file '%s'" % (fname,))
        raise
    vals = []
    it = iter(lut_file)
    # Skip the first line
    it.next()
    for l in it:
        row = [float(x) for x in l.strip().split()]
        if len(row) == 4:
            vals.append(row)
    return vals


class LookupTable(object):
    """Lookup-table representation.

    This corresponds to a L{vtk.vtkLookupTable}.
    @ivar vtk_lut: The corresponding L{vtk.vtkLookupTable}.
    """
    def __init__(self, source=None, range=(0, 1)):
        """Constructor.
        @param source: Specify the source from which to define the table, either
        as a filename, a table or a L{vtk.vtkLookupTable}. It is possible to
        omit this parameter, in which case a stock table is employed.
        @param range: Optionally specify range. Default will be (0, 1).
        """
        if isinstance(source, vtk.vtkObject):
            self.vtk_lut = source
            return
        vtk_lut = self.vtk_lut = vtk.vtkLookupTable()
        if isinstance(source, basestring):
            # Filename
            source = _read_lut_file(source, vtk_lut)
        elif source is None:
            source = _STOCK_LUT

        # Table
        vtk_lut.SetNumberOfTableValues(len(source))
        for i, val in enumerate(source):
            vtk_lut.SetTableValue(i, *val)
        # vtk_lut.Build()

        if range is not None:
            self.range = range

    
    def __eq__(self, rhs):
        this_lut, rhs_lut = self.vtk_lut, rhs.vtk_lut
        if (rhs_lut.GetNumberOfTableValues() !=
                this_lut.GetNumberOfTableValues()):
            return False
        for i in range(this_lut.GetNumberOfTableValues()):
            if rhs_lut.GetTableValue(i) != this_lut.GetTableValue(i):
                return False
        if rhs.range != self.range:
            return False

        return True


    def __get_range(self):
        """Get range."""
        return self.vtk_lut.GetRange()

    def __set_range(self, (rmin, rmax)):
        """Set range."""
        self.vtk_lut.SetRange(rmin, rmax)

    range = property(__get_range, __set_range)


def set_default_lut(lut):
    """Set the default lookup-table for actors which don't have one
    specified.
    @param lut: This may either be a L{LookupTable} object, or a parameter
    acceptable to this class.
    """
    global _default_lut
    if not isinstance(lut, LookupTable):
        lut = LookupTable(lut)
    _default_lut = lut

def get_default_lut():
    """Get the default lookup-table.
    """
    return _default_lut

set_default_lut(None)


class ActorBase(object):
    """Baseclass for Viper actors.
    
    This class corresponds to the actor concept in VTK.
    @ivar source: The relevant VTK source for this object. This attribute
    depends on the subclass.
    @ivar is_visible: Is the actor visible? Corresponds to
    vtk.vtkActor.Get/SetVisibility.
    @ivar color: The actor's color, specified as an RGB tuple.
    @ivar opacity: The color's opacity, from 0.0 (transparent) to 1.0.
    @ivar scalar_range: The possible range for scalar values. This range will
    be specific to the actor, independent of the lookup-table.
    @ivar lut: The associated L{lookup-table<LookupTable>} for translating
    scalar values into colors.
    @ivar vtk_actor: VTK actor.
    @ivar sig_modified: Emitted when the visual state is modified.
    @ivar _mapper: Polygon data mapper.
    @ivar _source: VTK data L{array<vtk.vtkDoubleArray>}.
    @ivar _refs: Put references to objects that are passed as void pointers to
    VTK here!
    """
    def __init__(self, source, wireframe=False, position=(0, 0, 0),
            lut=None, opacity=None, color=(0, 0, 0)):
        """Constructor.
        @param lut: Optionally, specify a L{lookup-table<LookupTable>}. If unspecified,
        the global default LUT is employed. You may also specify a type of
        parameter that is accepted by LookupTable's constructor here.
        @param color: Optionally, specify color. Default: black.
        """
        self.source = source
        actor = self.vtk_actor = vtk.vtkActor()
        actor.AddPosition(position)
        self._refs = []
        self.sig_modified = Signal()
        if opacity is not None:
            self.opacity = opacity
        self.color = color

        if wireframe:
            self.set_property("RepresentationToWireframe")

        mapper = self._mapper = vtk.vtkPolyDataMapper()
        self.vtk_actor.SetMapper(mapper)
        self.lut = lut


    def get_property(self, name):
        """Get a property of the associated VTK actor.
        
        Corresponds to vtk_actor.GetProperty().Get<Name>.
        """
        try: mthd = getattr(self.vtk_actor.GetProperty(), "Get" + name)
        except AttributeError:
            raise ValueError(name)
        return mthd()


    def set_property(self, name, *args):
        """Set a property of the associated VTK actor.
        
        Corresponds to vtk_actor.GetProperty().Set<Name>.
        """
        try: mthd = getattr(self.vtk_actor.GetProperty(), "Set" + name)
        except AttributeError:
            raise ValueError(name)
        mthd(*args)


    def __get_visible(self):
        return self.vtk_actor.GetVisibility()

    def __set_visible(self, shown):
        self.vtk_actor.SetVisibility(shown)

    is_visible = property(__get_visible, __set_visible)


    def __get_color(self):
        return self.get_property("Color")

    def __set_color(self, color):
        self.set_property("Color", *color)

    color = property(__get_color, __set_color)


    def __get_opacity(self):
        return self.get_property("Opacity")

    def __set_opacity(self, opacity):
        self.set_property("Opacity", opacity)

    opacity = property(__get_opacity, __set_opacity)


    def __get_scalar_range(self):
        return self._mapper.GetScalarRange()

    def __set_scalar_range(self, range):
        self._mapper.SetScalarRange(*range)

    scalar_range = property(__get_scalar_range, __set_scalar_range)


    def __get_lut(self):
        return self.__lut

    def __set_lut(self, lut):
        """Set the lookup-table.
        @param lut: The L{lookup-table<LookupTable>}. If None, the global
        default LUT is employed. You may also use a type of parameter that is
        accepted by LookupTable's constructor here.
        """
        if lut is None:
            self.__lut = _default_lut
        else:
            if not isinstance(lut, LookupTable):
                lut = LookupTable(lut)
            self.__lut = lut
        self._mapper.SetLookupTable(self.lut.vtk_lut)

    lut = property(__get_lut, __set_lut)


    def update_source(self, new_source=None):
        """Update the VTK source when its data changes implicitly.

        This can be used to either tell the actor to update its data source, or
        to set a new data source.
        @param new_source: Optionally specify new source data.
        """
        if new_source is not None:
            # Subclass must implement this
            self._set_source(new_source)
        else:
            self.source.Modified()


    def get_outline(self):
        """Get L{outline<vtk.vtkActor>} visualization."""
        raise NotImplementedError()


def CellArray(cells):
    """Create a L{numpy array<numpy.ndarray>} of cells.
    
    Each cell is a tuple of vertex IDs. This specialization ensures that IDs
    are of the type expected by VTK.
    """
    return numpy.array(cells, dtype=viper._util.get_vtk_idtype(), copy=False)

def CellZeros(shape):
    """Create a L{numpy array<numpy.ndarray>} of cells, populated with zeros.
    
    Each cell is a tuple of vertex IDs. This specialization ensures that IDs
    are of the type expected by VTK.
    """
    return numpy.zeros(shape, dtype=viper._util.get_vtk_idtype())


def _create_grid(vertices, cells, refs):
    """Create L{unstructured grid<vtk.vtkUnstructuredGrid>} from vertices and
    cells."""
    ita = vtk.vtkIdTypeArray()
    if cells.dtype != viper.get_vtk_idtype():
        raise ValueError("Wrong cell data type, size is %d, must be %d" %
                (cells.itemsize, ita.GetDataTypeSize()))

    cell_dim = cells.shape[1]
    if cell_dim == 2:
        celltype = 3
    elif cell_dim == 3:
        celltype = 5
    elif cell_dim == 4:
        celltype = 10

    cellsize = cell_dim + 1

    nl = cells.copy()
    cl = numpy.zeros((vertices.shape[0], 3), dtype='d')
    cl[:, :vertices.shape[1]] = vertices
    
    # Create a cell array, stored in Python.numpy
    inttype = nl.dtype
    ncells = len(nl)
    cells = numpy.zeros((ncells, cellsize), dtype=inttype)
    cells[:, 1:cellsize] = nl
    cells[:, 0] = cellsize - 1 
    
    a = numpy.ravel(cells)
    # Important: we must keep references since VTK doesn't!
    refs.append(a)
    ita.SetVoidArray(a, len(a), 1)
    ca = vtk.vtkCellArray()
    ca.SetCells(ncells, ita)
    
    # Create some points
    npoints = len(cl)
    pts = numpy.ravel(cl)
    
    pa = vtk.vtkDoubleArray()
    pa.SetNumberOfComponents(3)
    # Important: we must keep references since VTK doesn't!
    refs.append(pts)
    pa.SetVoidArray(pts, npoints*3, 1)
    
    v = vtk.vtkPoints()
    v.SetNumberOfPoints(npoints)
    v.SetData(pa)
    
    grid = vtk.vtkUnstructuredGrid()
    grid.SetPoints(v)
    grid.SetCells(celltype, ca)

    return grid


class ILabelable(object):
    """Interface for objects that support labeling.
    @ivar is_point_labels_enabled: Are point labels enabled?
    @ivar is_cell_labels_enabled: Are point labels enabled?
    @ivar sig_point_labels_enabled: Emitted when is_point_labels_enabled changes.
    @ivar sig_cell_labels_enabled: Emitted when is_cell_labels_enabled changes.
    """
    def get_point_labels_actor(self):
        """Get actor for point labels."""


    def get_cell_labels_actor(self):
        """Get actor for cell labels."""
        

class _Labelable(ILabelable):
    """Mixin implementation of ILabelable)."""
    def __init__(self):
        self.sig_point_labels_enabled, self.sig_cell_labels_enabled = (
                Signal(), Signal())
        self.__is_plabels_enabled = False
        self.__is_clabels_enabled = False


    def __get_plabels(self):
        return self.__is_plabels_enabled

    def __set_plabels(self, enable):
        self.__is_plabels_enabled = enable
        self.sig_point_labels_enabled(enable)

    is_point_labels_enabled = property(__get_plabels, __set_plabels)


    def __get_clabels(self):
        return self.__is_clabels_enabled

    def __set_clabels(self, enable):
        self.__is_clabels_enabled = enable
        self.sig_cell_labels_enabled(enable)

    is_cell_labels_enabled = property(__get_clabels, __set_clabels)


    def get_point_labels_actor(self):
        """Get actor for point labels."""
        ids = vtk.vtkIdFilter()
        ids.PointIdsOn()
        ids.FieldDataOn()
        self._set_label_id_input(ids)

        ldm = vtk.vtkLabeledDataMapper()
        ldm.SetLabelFormat("%g")
        ldm.SetInputConnection(ids.GetOutputPort())
        ldm.SetLabelModeToLabelFieldData()
        ldm.GetLabelTextProperty().SetColor(0, 0, 0)
        ldm.GetLabelTextProperty().BoldOn()
        point_labels = vtk.vtkActor2D()
        point_labels.SetMapper(ldm)
        return point_labels


    def get_cell_labels_actor(self):
        """Get actor for cell labels."""
        ids = vtk.vtkIdFilter()
        self._set_label_id_input(ids)
        ids.CellIdsOn()
        ids.FieldDataOn()

        ldm = vtk.vtkLabeledDataMapper()
        ldm.SetLabelFormat("%g")
        ldm.SetInputConnection(ids.GetOutputPort())
        ldm.SetLabelModeToLabelFieldData()
        ldm.GetLabelTextProperty().SetColor(0, 0, 0)
        ldm.GetLabelTextProperty().BoldOn()
        cell_labels = vtk.vtkActor2D()
        cell_labels.SetMapper(ldm)
        return cell_labels


class _MeshBase(ActorBase, _Labelable):
    """Base for mesh types.
    @ivar _grid: Associated VTK grid.
    @ivar _scalars: Scalars.
    @ivar _pipeline_start: The start of the processing pipeline.
    """
    def __init__(self, vertices, cells, scalars, actor_args=(), actor_kwds={}):
        _Labelable.__init__(self)
        """Constructor.
        @param vertices: L{Array<numpy.ndarray>} of vertices.
        @param cells: L{Array<numpy.ndarray>} of cells.
        """

        source = vtk.vtkDoubleArray()
        # Make sure to keep a reference
        self._scalars = scalars
        source.SetVoidArray(scalars, len(scalars), 1)
        ActorBase.__init__(self, source, *actor_args, **actor_kwds)

        self.__outline = None
        grid = self._grid = _create_grid(vertices, cells, self._refs)
        grid.GetPointData().SetScalars(self.source)
        

        # Shading
        extract = self._pipeline_start = vtk.vtkGeometryFilter()
        extract.SetInput(self._grid)
        extract.GetOutput().ReleaseDataFlagOn()
        normals = vtk.vtkPolyDataNormals()
        normals.SetInputConnection(extract.GetOutputPort())
        self._mapper.SetInputConnection(normals.GetOutputPort())


    def get_outline(self):
        """Get outline."""
        filter = vtk.vtkOutlineFilter()
        filter.SetInput(self._grid)
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(filter.GetOutputPort())
        outline = vtk.vtkActor()
        outline.SetMapper(mapper)
        outline.GetProperty().SetColor(1, 0, 0)
        return outline


    def _set_source(self, scalars):
        """Set the VTK source, called by baseclass.
        """
        self._scalars = scalars
        self.source.SetVoidArray(scalars, len(scalars), 1)

    
    def _set_label_id_input(self, ids):
        ids.SetInput(self._grid)


class Mesh(_MeshBase):
    """Basic mesh representation.
    """
    def __init__(self, vertices, cells, actor_args=(), actor_kwds={}):
        actor_kwds["wireframe"] = True
        _MeshBase.__init__(self, vertices, cells, numpy.zeros(len(vertices),
            dtype='d'), actor_args, actor_kwds)


def _create_cut_filter(cells, origin, normal):
    """Create a cutplane filter."""
    if cells.shape[1] != 4:
        raise ValueError("Can only cut scalar 3 dimensional plots")
    
    if not (isinstance(origin, (numpy.ndarray, tuple, list)) and 
            len(origin) == 3):
        raise ValueError("Provide a tuple, list or numpy array of length 3 for "
                "cutplane origin")
    
    if not (isinstance(normal, (numpy.ndarray, tuple, list)) and 
            len(normal) == 3):
        raise ValueError("Provide a tuple, list or numpy array of length 3 for "
                "cutplane normal")

    # Define the cut plane
    plane = vtk.vtkPlane()
    plane.SetOrigin(origin)
    plane.SetNormal(normal)

    # Initialize the Cutter filter
    cutter = vtk.vtkCutter()
    cutter.SetCutFunction(plane)
    return cutter



class ScalarField(_MeshBase):
    """ A scalar field over a mesh.
    
    Basically this extends the basic mesh with scalars.

    The implementation difference is that we supply a variable set of scalars
    and and a variable scalar range, whereas the mesh only supplies zero-valued
    scalars.

    It is possible to visualize only a cutplane of the field.
    @ivar scalars: The scalars.
    """
    def __init__(self, vertices, cells, scalars, scalar_range=None,
            cutplane_origin=None, cutplane_normal=(0, 0, 1), displacement=None,
            actor_args=(), actor_kwds={}):
        """Constructor.
        @param vertices: Mesh vertics.
        @param cells: Mesh L{cells<numpy.ndarray>}.
        @param scalars: Scalar value per vertex.
        @param scalar_range: Optionally, specify the scalar range. If
        unspecified, it will correspond to the range of your scalars.
        @param cutplane_origin: Optionally define a cutplane's origin.
        @param cutplane_normal: If a cutplane has been specified, optionally
        specify its normal.
        @param displacement: Optionally specify a displacement vector field.
        """
        _MeshBase.__init__(self, vertices, cells, scalars, actor_args=actor_args,
                actor_kwds=actor_kwds)

        mapper = self._mapper
        if scalar_range is None:
            # Auto
            self.scalar_range = (numpy.min(scalars), numpy.max(scalars))
        else:
            self.scalar_range = scalar_range

        disp_filter = None
        if displacement is not None:
            if displacement.shape != (len(scalars), 3):
                raise ValueError(
                        "Wrong shape of vector displacement field (must be 3D)")
            disp_filter = self.__create_disp_filter(displacement)
            disp_filter.SetInput(self._grid)
            self._pipeline_start.SetInputConnection(disp_filter.GetOutputPort())

        if cutplane_origin is not None:
            cut_filter = _create_cut_filter(cells, cutplane_origin,
                    cutplane_normal)
            if disp_filter is None:
                cut_filter.SetInput(self._grid)
            else:
                cut_filter.SetInputConnection(disp_filter.GetOutputPort())
            self._pipeline_start.SetInputConnection(cut_filter.GetOutputPort())


    def update_displacement(self):
        """Update the displacement filter."""
        self.__disp_source.Modified()

    @property
    def scalars(self):
        return self._scalars


    def __create_disp_filter(self, disp):
        """Create a displacement filter.
        
        The displacement is achieved by warping the grid with a set of vectors.
        @param disp: Displacement vectors.
        @return: L{Displacement filter<vtk.vtkWarpVector>}.
        """
        source = self.__disp_source = vtk.vtkDoubleArray()
        source.SetNumberOfComponents(3)
        self._refs.append(disp)
        source.SetVoidArray(disp, disp.shape[0]*disp.shape[1], 1)
        self._grid.GetPointData().SetVectors(source)

        warp = vtk.vtkWarpVector()
        warp.SetScaleFactor(1.0)
        return warp


def _create_outline(grid):
    """Create an outline pipeline, represented by a L{vtk.vtkPolyDataMapper}."""
    outline = vtk.vtkOutlineFilter()
    outline.SetInput(grid)
    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputConnection(outline.GetOutputPort())
    return mapper


class VectorField(ActorBase):
    """A vector field.
    @ivar directions: Directions.
    """
    def __init__(self, vertices, directions, cells=None, actor_args=(),
            actor_kwds={}):
        self.directions = directions
        source = vtk.vtkDoubleArray()
        source.SetNumberOfComponents(3)
        source.SetVoidArray(directions, len(numpy.ravel(directions)), 1)

        ActorBase.__init__(self, source, *actor_args, **actor_kwds)

        vmin, vmax = self.__comp_range(directions)
        glyph = self.__create_glyph(vertices, source, cells, vmax)

        actor = self.vtk_actor
        mapper = self._mapper
        mapper.SetInput(glyph.GetOutput())
        mapper.SetScalarRange(vmin, vmax)

    def _set_source(self, directions):
        """Set the VTK source, called by baseclass.
        """
        self.directions = directions
        self.source.SetNumberOfComponents(3)
        self.source.SetVoidArray(directions, len(numpy.ravel(directions)), 1)

    def __create_glyph(self, vertices, source, cells, vmax):
        glyph = vtk.vtkGlyph3D()
        glyph.SetColorModeToColorByVector()
        glyph.SetScaleModeToScaleByVector()

        n, m = vertices.shape
        vertices_3d = numpy.zeros((n, 3), dtype='d')
        vertices_3d[:,:m] = vertices
        '''
        directions_3d = numpy.zeros((n, 3), dtype='d')
        if directions.shape[1] > m:
            # Directions is 3d already
            m = directions.shape[1]
        directions_3d[:,:m] = directions
        '''

        m = 3
        
        vtkpoints = vtk.vtkDoubleArray()
        vtkpoints.SetNumberOfComponents(m)
        self._refs.append(vertices_3d)
        vtkpoints.SetVoidArray(vertices_3d, n*m, m)
        _points = vtk.vtkPoints()
        _points.SetData(vtkpoints)

        '''
        vtkdirections = vtk.vtkDoubleArray()
        vtkdirections.SetNumberOfComponents(m)
        self._refs.append(directions_3d)
        vtkdirections.SetVoidArray(directions_3d, n*m, m)
        '''

        data = vtk.vtkPolyData()
        data.SetPoints(_points)
        data.GetPointData().SetVectors(source)
        glyph.SetInput(data)
        glyph.SetVectorModeToUseVector()

        arrow = vtk.vtkArrowSource()
        arrow.SetTipRadius(0.15)
        arrow.SetTipLength(0.15)
        arrow.SetShaftRadius(0.05)
        glyph.SetSource(arrow.GetOutput())

        # Auto-scale vectors,
        nnz, nsd = vertices.shape
        if cells is not None:
            grid = _create_grid(vertices, cells, self._refs)
            a = _create_outline(grid).GetBounds()
        else:
            # Assume unit cube
            a = (0, 1, 0, 1, 0, 1)
        D = math.sqrt(reduce(operator.add,[(a[i]-a[i-1])**2 for i in
            xrange(1,6,2)]))
        sf = 2.0
        sf = sf/vmax*D/(nnz**(1.0/nsd))
        glyph.SetScaleFactor(sf)

        return glyph

    def __comp_range(self, directions):
        norms = [numpy.linalg.norm(v) for v in directions]
        vmin, vmax = min(norms), max(norms)
        if abs(vmax) < 1e-16:
            vmax = 1.0
        return vmin, vmax


class Sphere(ActorBase, _Labelable):
    """Sphere representation.
    """
    def __init__(self, point, radius, theta_res=8, phi_res=6,
            actor_args=(), actor_kwds={}):
        _Labelable.__init__(self)

        sphere = self.__sphere = vtk.vtkSphereSource()
        sphere.SetThetaResolution(theta_res)
        sphere.SetPhiResolution(phi_res)
        sphere.SetRadius(radius)
        sphere.SetCenter(*point)
        ActorBase.__init__(self, sphere, *actor_args, **actor_kwds)

        mapper = self._mapper
        mapper.SetInputConnection(sphere.GetOutputPort())


    def _set_label_id_input(self, ids):
        ids.SetInputConnection(self.__sphere.GetOutputPort())


class Polygon(ActorBase, _Labelable):
    """Polygon representation.
    """
    def __init__(self, vertices, actor_args=(), actor_kwds={}):
        ActorBase.__init__(self, None, *actor_args, **actor_kwds)
        _Labelable.__init__(self)

        is_2d = len(vertices[0]) == 2
        num_vertices = len(vertices)
        vtk_points = vtk.vtkPoints()
        vtk_points.SetNumberOfPoints(num_vertices)

        data = numpy.zeros((num_vertices, 3))
        data[:,:len(vertices[0])] = numpy.array(vertices)
        self._refs.append(data)
        pa = vtk.vtkDoubleArray()
        pa.SetNumberOfComponents(3)
        pa.SetVoidArray(data, num_vertices*3, 1)
        vtk_points.SetData(pa)

        line = vtk.vtkPolyLine()
        line.GetPointIds().SetNumberOfIds(num_vertices)
        for i in xrange(num_vertices):
            line.GetPointIds().SetId(i, i)
        grid = vtk.vtkUnstructuredGrid()
        grid.Allocate(1, 1)
        grid.InsertNextCell(line.GetCellType(),
                            line.GetPointIds())
        grid.SetPoints(vtk_points)

        self._grid = grid

        extract = vtk.vtkGeometryFilter()
        extract.SetInput(grid)
        extract.GetOutput().ReleaseDataFlagOn()
        normals = vtk.vtkPolyDataNormals()
        normals.SetInputConnection(extract.GetOutputPort())

        mapper = self._mapper
        mapper.SetInputConnection(normals.GetOutputPort())
        self.set_property("LineWidth", 1)

    def _set_label_id_input(self, ids):
        ids.SetInput(self._grid)


class _Scalarbar(object):
    """Scalarbar representation.

    A scalarbar visualizes the scalar range of its associated actor and the
    corresponding color spectrum.

    The scalarbar's LUT is filled in like the actor's, and its range is set to
    that of the actor (which may just mirror that of its LUT).
    @ivar lut: The L{lookup table<vtk.vtkLookupTable>}.
    @ivar actor: Eventual associated actor.
    @ivar should_show: Should the scalarbar be shown?
    @ivar is_visible: Is the scalarbar currently visible? For this to hold,
    should_show must be true and at least one actor must be associated.
    """
    def __init__(self, vip, pos):
        self.__vip = vip

        actor = self.__bar_actor = vtk.vtkScalarBarActor()
        actor.VisibilityOff()
        actor.GetPositionCoordinate().SetCoordinateSystemToNormalizedViewport()
        actor.GetTitleTextProperty().SetColor(0,0,0)
        actor.GetLabelTextProperty().SetColor(0,0,0)

        if pos == vip.POS_N:
            actor.GetPositionCoordinate().SetValue (0.1,0.91)
            actor.SetOrientationToHorizontal()
            actor.SetWidth(0.8)
            actor.SetHeight(0.12)
        elif pos == vip.POS_S:
            actor.GetPositionCoordinate().SetValue (0.1,0.01)
            actor.SetOrientationToHorizontal()
            actor.SetWidth(0.8)
            actor.SetHeight(0.12)
        elif pos == vip.POS_E:
            actor.GetPositionCoordinate().SetValue (0.01,0.15)
            actor.SetOrientationToVertical()
            actor.SetWidth(0.12)
            actor.SetHeight(0.85)
        elif pos == vip.POS_W:
            actor.GetPositionCoordinate().SetValue (0.81,0.15)
            actor.SetOrientationToVertical()
            actor.SetWidth(0.12)
            actor.SetHeight(0.85)

        vip.renderer.AddViewProp(actor)

        self.__actor = None
        self.__should_show = False

    @property
    def lut(self):
        return LookupTable(self.__bar_actor.GetLookupTable())


    def __get_actor(self):
        return self.__actor

    def __set_actor(self, actor):
        if self.__actor is not None:
            self.__actor.vtk_actor.GetMapper().RemoveObserver(self.__mapper_obs)
        self.__actor = actor
        if actor is not None:
            self.__update_lut()
            self.__mapper_obs = actor.vtk_actor.GetMapper().AddObserver(
                    "ModifiedEvent", self.__update_lut)
        self.__update_visibility()

    actor = property(__get_actor, __set_actor)


    @property
    def is_visible(self):
        return self.should_show and self.actor is not None


    def __get_show(self):
        return self.__should_show

    def __set_show(self, show):
        self.__should_show = show
        self.__update_visibility()

    should_show = property(__get_show, __set_show)


    def __update_visibility(self):
        bar_actor = self.__bar_actor
        if self.__should_show and self.actor is not None:
            bar_actor.VisibilityOn()
        else:
            bar_actor.VisibilityOff()

    def __update_lut(self, *args):
        bar_actor = self.__bar_actor
        if self.__actor is not None:
            vtk_actor = self.__actor.vtk_actor
            mapper = self.__actor._mapper
            lut = self.__actor.lut.vtk_lut
            if bar_actor.GetLookupTable() is None:
                bar_actor.SetLookupTable(vtk.vtkLookupTable())
            bar_lut = bar_actor.GetLookupTable()
            bar_lut.SetNumberOfTableValues(lut.GetNumberOfTableValues())
            for i in range(lut.GetNumberOfTableValues()):
                bar_lut.SetTableValue(i, lut.GetTableValue(i))
            if mapper.GetUseLookupTableScalarRange():
                bar_lut.SetRange(lut.GetRange())
            else:
                bar_lut.SetRange(mapper.GetScalarRange())
        else:
            bar_actor.SetLookupTable(None)


lutdir = os.path.join(os.path.split(__file__)[0], "data")

class Viper(object):
    """Our VTK plotter.
    
    Viper provides a simple VTK framework for plotting L{actors<ActorBase>}.
    The actual plotting takes place by rendering each actor in a L{VTK
    renderer<vtk.vtkRenderer>}.

    @ivar renderer: The VTK renderer.
    @ivar actors: List of associated L{actors<ActorBase>}.
    @ivar outline_enabled: Are outlines turned on?
    @ivar sig_modified: Emitted when the visual state is modified.
    """
    POS_N, POS_S, POS_E, POS_W = range(4)

    __cam_confspec = {"position": "float_list(min=3, max=3)",
                "focal-point": "float_list(min=3, max=3)",
                "view-up": "float_list(min=3, max=3)",
                "clipping-range": "float_list(min=2, max=2)",
                "parallel-projection": "boolean",
                "view-angle": "float",
                "window-center": "float_list(min=2, max=2)",
                }

    def __init__(self, actors=None):
        """Create plotter.
        @param actors: Optionally specify objects to draw.
        """
        self.sig_modified = viper._signal.Signal()
        self.__initcommon()
        self._construct_renderer()
        self.__scalarbars = [_Scalarbar(self, p) for p in range(4)]
        self.__is_plabels_enabled = self.__is_clabels_enabled = False
        self.__plabels = {}
        self.__clabels = {}

        if actors is None:
            actors = []
        if not isinstance(actors, (list, tuple)):
            actors = [actors]
        for actor in actors:
            self.add_actor(actor)

        self.reset_camera()

    def add_actor(self, actor):
        """Add L{actor<ActorBase>} to plot."""
        self.renderer.AddViewProp(actor.vtk_actor)
        self.actors.append(actor)
        if self.outline_enabled:
            self.__add_outline(actor)
        if isinstance(actor, ILabelable):
            actor.sig_point_labels_enabled.connect(self.__slot_plabels_enabled,
                    defKwds={"actor": actor})
            actor.sig_cell_labels_enabled.connect(self.__slot_clabels_enabled,
                    defKwds={"actor": actor})
            self.__enable_plabels(actor, actor.is_point_labels_enabled)
            self.__enable_clabels(actor, actor.is_cell_labels_enabled)


    def remove_actor(self, actor):
        """Remove an actor."""
        # Remove eventual outline
        if actor in self.__outlines:
            outline = self.__outlines.pop(actor)
            self.renderer.RemoveActor(outline)
        for scalarbar in self.__scalarbars:
            if actor is scalarbar.actor:
                scalarbar.actor = None
        self.actors.remove(actor)
        self.renderer.RemoveActor(actor.vtk_actor)

        if isinstance(actor, ILabelable):
            actor.sig_point_labels_enabled.disconnect(self.__slot_plabels_enabled)
            actor.sig_cell_labels_enabled.disconnect(self.__slot_clabels_enabled)
            self.__enable_plabels(actor, False)
            self.__enable_clabels(actor, False)


    def __slot_plabels_enabled(self, enable, actor):
        self.__enable_plabels(actor, enable)

    def __enable_plabels(self, actor, enable):
        ren = self.renderer
        if enable:
            plabels = actor.get_point_labels_actor()
            assert actor not in self.__plabels
            self.__plabels[actor] = plabels
            ren.AddViewProp(plabels)
        else:
            try: ren.RemoveActor(self.__plabels.pop(actor))
            except KeyError: pass
            

    def __slot_clabels_enabled(self, enable, actor):
        self.__enable_clabels(actor, enable)

    def __enable_clabels(self, actor, enable):
        ren = self.renderer
        if enable:
            clabels = actor.get_cell_labels_actor()
            assert actor not in self.__clabels
            self.__clabels[actor] = clabels
            ren.AddViewProp(clabels)
        else:
            try: ren.RemoveActor(self.__clabels.pop(actor))
            except KeyError: pass

    
    def get_vtk_actors(self):
        """Get VTK L{actors<vtk.vtkActor>}."""
        actors = []
        coll = self.renderer.GetViewProps()
        if coll is None:
            # It happens that None is returned!
            return actors
        for i in range(coll.GetNumberOfItems()):
            actor = vtk.vtkActor.SafeDownCast(coll.GetItemAsObject(i))
            if actor is not None:
                actors.append(actor)
        return actors


    def get_vtk_actors2d(self):
        """Get VTK L{2D actors<vtk.vtkActor2D>}."""
        actors = []
        coll = self.renderer.GetViewProps()
        if coll is None:
            # It happens that None is returned!
            return actors
        for i in range(coll.GetNumberOfItems()):
            actor = vtk.vtkActor2D.SafeDownCast(coll.GetItemAsObject(i))
            if actor is not None:
                actors.append(actor)
        return actors


    def clear(self):
        """Remove all actors."""
        for actor in self.actors[:]:
            self.remove_actor(actor)

    def toggle_outline(self):
        """Toggle outlines around actors."""
        if not self.outline_enabled:
            for actor in self.actors:
                self.__add_outline(actor)
        else:
            for outline in self.__outlines.values():
                self.renderer.RemoveActor(outline)
            self.__outlines = {}

        self.outline_enabled = not self.outline_enabled


    def reset_camera(self, center=True, clipping_range=True):
        """Reset camera.
        @param center: Center camera with regard to actors?
        @param clipping_range: Reset clipping range to encompass actors?
        """
        ren = self.renderer
        if center:
            ren.ResetCamera()
        if clipping_range:
            ren.ResetCameraClippingRange()

    def save_camera(self, filename):
        """Save camera state to file.

        Camera state is defined as the parameters position, focal point and
        view-up vector. Judging from VTK documentation, these are what matter.
        """
        cam = self.renderer.GetActiveCamera()
        conf = configobj.ConfigObj(filename, configspec=self.__cam_confspec)

        cam = self.renderer.GetActiveCamera()
        conf["position"] = list(cam.GetPosition())
        conf["focal-point"] = list(cam.GetFocalPoint())
        conf["view-up"] = list(cam.GetViewUp())
        conf["clipping-range"] = list(cam.GetClippingRange())
        conf["parallel-projection"] = bool(cam.GetParallelProjection())
        conf["view-angle"] = cam.GetViewAngle()
        conf["window-center"] = list(cam.GetWindowCenter())
        val = conf.validate(validate.Validator())
        assert val == True, val
        conf.write()

    def load_camera(self, filename):
        """Load camera state from file.
        
        The file should be created earlier by save_camera.
        """
        conf = configobj.ConfigObj(filename, configspec=self.__cam_confspec)
        val = conf.validate(validate.Validator())
        if val != True:
            raise Exception("Invalid camera state file")
        cam = self.renderer.GetActiveCamera()
        cam.SetPosition(conf["position"])
        cam.SetFocalPoint(conf["focal-point"])
        cam.SetViewUp(conf["view-up"])
        cam.SetClippingRange(conf["clipping-range"])
        cam.SetParallelProjection(conf["parallel-projection"])
        cam.SetViewAngle(conf["view-angle"])
        cam.SetWindowCenter(*conf["window-center"])

        
    def __initcommon(self):
        self.actors = []
        self.outline_enabled = False
        self.__outlines = {}
        self.renderer = self.ren = None

        '''
        # paramaters to control camera movement, default is no movement
        self.elevator = 0
        self.elevator_sign = 1
        self.set_camera_movement()
        '''


    def get_scalarbar(self, pos):
        """Get L{scalarbar<_Scalarbar>} for a certain position."""
        try: return self.__scalarbars[pos]
        except IndexError: raise ValueError(pos)


    def azimuth(self, angle):
        """Rotate the active camera about the view up vector centered at the
        focal point. Note that the view up vector is not necessarily
        perpendicular to the direction of projection.
        """
        self.renderer.GetActiveCamera().Azimuth(angle)

    def elevate(self, x):
        self.renderer.GetActiveCamera().Elevation(x)

    def dolly(self, x):
        self.renderer.GetActiveCamera().Dolly(x)

    def set_viewangle(self, i):
        self.renderer.GetActiveCamera().SetViewAngle(i)


    def _construct_renderer(self):
        """ Construct the VTK renderer. """
        ren = self.renderer = self.ren = vtk.vtkRenderer()
        ren.GlobalWarningDisplayOff()
        ren.SetBackground(1, 1, 1)
        
        return ren


    def __add_outline(self, actor):
        outline = actor.get_outline()
        self.__outlines[actor] = outline
        self.renderer.AddViewProp(outline)


class VtkWindow(WidgetBase):
    """Standalone VTK window.
    
    Use this when you don't have/need a full-blown GUI toolkit.
    @ivar vtk_window: The L{VTK window<vtk.vtkRenderWindow>}.
    """
    def __init__(self, vip, title="FEniCS Viper"):
        WidgetBase.__init__(self, vip)
        # Configure logging handler, so log gets printed. The viper library
        # itself has no initial handler, since this should be up to the
        # application.
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
        logging.getLogger("Viper").addHandler(handler)

        window = self.vtk_window = vtk.vtkRenderWindow()
        window.SetWindowName(title)
        window.SetSize(800, 600)
        window.AddRenderer(vip.renderer)
        vip.sig_modified.connect(self.__render)

        iren = self.__iren = vtk.vtkRenderWindowInteractor()
        iren.SetRenderWindow(window)
        style = vtk.vtkInteractorStyleSwitch()
        style.SetCurrentStyleToTrackballCamera()
        iren.SetInteractorStyle(style)
        iren.AddObserver("KeyPressEvent", self.__key_press_methods)
        iren.Initialize()

    def interactive(self, dbug=False):
        _logger.info("Plot active, press 'q' to continue.")
        _logger.debug("Starting mainloop")
        self.__iren.Start()
        _logger.debug("Exited mainloop")

    def render(self):
        self.vtk_window.Render()

    def __key_press_methods(self, obj, event):
        """React to keypress."""
        vip = self.viper
        key = obj.GetKeyCode()
        _logger.debug("Keypress: '%s'" % key)
        _logger.info("Keypress: '%s'" % key)
        if key == "o":
            vip.toggle_outline()
            self.vtk_window.Render()
        elif key == "i":
            _logger.info("Writing a simulation frame to a 'png' file")
            self.write_png()
        elif key == "v":
            _logger.info("Writing data to a 'vtk' file")
            self.write_vtk()
        elif key == "m":
            _logger.info("Writing a simulation frame to an 'eps' file")
            self.write_ps()
        elif key == "n":
            _logger.info("Writing a simulation frame to a 'pdf' file")
            self.write_ps(format="pdf")
        elif key == "h":
            self.__display_keybindings()
        elif key == 'e':
            raise SystemExit
        else:
            _logger.debug("Unhandled key '%s'" % key)

    def __render(self):
        self.vtk_window.Render()

    def __display_keybindings(self):
        print """\
Keybindings:

Viper specific:
* Keypress v: write data to file in vtk format
* Keypress i: write image to file in png format
* Keypress m: write image to file in eps format
* Keypress n: write image to file in pdf format
* Keypress o: add outline 
* Keypress h: display this message

Inherited from vtk:
* Keypress j / Keypress t: toggle between joystick (position sensitive) and trackball (motion sensitive) styles. In joystick style, motion occurs continuously as long as a mouse button is pressed. In trackball style, motion occurs when the mouse button is pressed and the mouse pointer moves.

* Keypress c / Keypress a: toggle between camera and actor modes. In camera mode, mouse events affect the camera position and focal point. In actor mode, mouse events affect the actor that is under the mouse pointer.

* Button 1: rotate the camera around its focal point (if camera mode) or rotate the actor around its origin (if actor mode). The rotation is in the direction defined from the center of the renderer's viewport towards the mouse position. In joystick mode, the magnitude of the rotation is determined by the distance the mouse is from the center of the render window.

* Button 2: pan the camera (if camera mode) or translate the actor (if actor mode). In joystick mode, the direction of pan or translation is from the center of the viewport towards the mouse position. In trackball mode, the direction of motion is the direction the mouse moves. (Note: with 2-button mice, pan is defined as <Shift>-Button 1.)

* Button 3: zoom the camera (if camera mode) or scale the actor (if actor mode). Zoom in/increase scale if the mouse position is in the top half of the viewport; zoom out/decrease scale if the mouse position is in the bottom half. In joystick mode, the amount of zoom is controlled by the distance of the mouse pointer from the horizontal centerline of the window.

* Keypress 3: toggle the render window into and out of stereo mode. By default, red-blue stereo pairs are created. Some systems support Crystal Eyes LCD stereo glasses; you have to invoke SetStereoTypeToCrystalEyes() on the rendering window.

* Keypress e: exit the application.

* Keypress f: fly to the picked point

* Keypress p: perform a pick operation. The render window interactor has an internal instance of vtkCellPicker that it uses to pick.

* Keypress q: quit application

* Keypress r: reset the camera view along the current view direction. Centers the actors and moves the camera so that all actors are visible.

* Keypress s: modify the representation of all actors so that they are surfaces.

* Keypress u: invoke the user-defined function. Typically, this keypress will bring up an interactor that you can type commands in.

* Keypress w: modify the representation of all actors so that they are wireframe.
"""


class PngWriter(object):
    """Capture Viper visualizations as PNG images."""
    def __init__(self, vip):
        self.__viper = vip
        source = self.__source = vtk.vtkRenderLargeImage()
        source.SetInput(vip.renderer)
        source.SetMagnification(1)
        writer = self.__writer = vtk.vtkPNGWriter()
        writer.SetInputConnection(source.GetOutputPort())

    def capture_frame(self, filename):
        self.__viper.renderer.Render()
        self.__source.Update()
        self.__writer.SetFileName(filename)
        # Otherwise the writer might not sync with the input
        self.__source.Modified()
        self.__writer.Write()
        assert os.path.exists(filename)


class NoFramesError(ViperError):
    """No frames have been recorded."""

class MovieMaker(object):
    """Turn Viper visualizations into movies.

    Viper frames are captured on demand as image files, and finally turned into
    a movie.

    Upon initialization, an exception will be raised if a suitable movie
    encoder can't be found. Currently, only mencoder is supported.
    @ivar filename: Movie filename.
    @ivar fps: Frames per second
    """
    def __init__(self, vip, filename, fps=25):
        self.filename = filename
        self.fps = fps
        self.__writer = PngWriter(vip)
        self.__frames = []
        self.__tmpdir = tempfile.mkdtemp()

        try: viper._util.resolve_path("mencoder")
        except viper._util.NotFound:
            raise ViperError("Can't find any suitable movie encoder")

    def capture_frame(self):
        """Capture the current visual state as a frame."""
        fd, fname = tempfile.mkstemp(dir=self.__tmpdir, suffix=".png")
        os.close(fd)
        self.__writer.capture_frame(fname)
        self.__frames.append(fname)

    def make_movie(self, close=True):
        """Make the movie from the individual frames.
        
        This function blocks until the movie encoder has finished.
        @param close: Also close?
        @raise NoFramesError: No frames recorded.
        """
        if not self.__frames:
            raise NoFramesError
        fnames = ",".join(self.__frames)
        import sys
        args = ["mencoder", "mf://%s" % fnames, "-mf", "fps=%d" % self.fps,
                "-o", self.filename, "-ovc", "lavc", "-lavcopts",
                "vcodec=mpeg4"]
        #command = 'mencoder mf://%s -mf fps=%d -o %s -ovc lavc -lavcopts vcodec=mpeg4:mbd=2:mv0:trell:v4mv:cbp:last_pred=3:predia=2:dia=2:vmax_b_frames=2:vb_strategy=1:precmp=2:cmp=2:subcmp=2:preme=2:qns=2' % (files, self.fps, self.output)
        _logger.debug("Encoding with mencoder: '%s'" % " ".join(args[1:]))
        # Make sure that eventual debug output is printed before mencoder's
        # messages
        sys.stderr.flush()
        try: subprocess.check_call(args)
        except subprocess.CalledProcessError, err:
            raise ViperError("mencoder failed: %s" % unicode(err))
        if close:
            self.close()

    def close(self):
        """Release resources."""
        self.__frames = []
        shutil.rmtree(self.__tmpdir)
