"""Test _viper module.
"""
from viper.tests._common import *

import numpy
import os.path
import vtk
import subprocess
import sys

import viper._viper


_on_windows = sys.platform.startswith("win")

_vertices1 = numpy.array([[0.0, 0.0, 0.0],
                        [1.0, 0.0, 0.0],
                        [0.0, 1.0, 0.0],
                        [1.0, 1.0, 0.0],
                        [2.0, 0.0, 0.0],
                        [0.0, 2.0, 0.0],
                        [2.0, 1.0, 0.0],
                        [1.0, 2.0, 0.0],
                        [2.0, 2.0, 0.0]])
_cells1 = viper.CellArray([[0, 1, 2], [1, 2, 3], [1, 4, 3], [4, 3, 6], [2, 3, 5],
                        [3, 5, 7], [3, 6, 7], [6, 7, 8]])
_scalars1 = numpy.linspace(0, 1, len(_vertices1))


_vertices2 = numpy.array([[0.0, 0.0, 0.0],
                        [1.0, 0.0, 0.0],
                        [0.0, 1.0, 0.0],
                        [1.0, 1.0, 0.0],
                        [2.0, 0.0, 0.0],
                        [0.0, 2.0, 0.0],
                        [2.0, 1.0, 0.0],
                        [1.0, 2.0, 0.0],
                        [2.0, 2.0, 0.0]])
_scalars2 = _vertices2.copy()
_cells2 = viper.CellArray([[0, 1, 2], [1, 2, 3], [1, 4, 3], [4, 3, 6], [2, 3, 5],
                        [3, 5, 7], [3, 6, 7], [6, 7, 8]])

_vertices3 = numpy.array([[0, 0, 0], [1, 0, 0],
                        [0, 1, 0], [0, 0, 1]])
_scalars3 = numpy.linspace(0, 1, len(_vertices3))
_cells3 = viper.CellArray([(0, 1, 2, 3),])

_lut_fname = os.path.realpath(os.path.join(os.path.dirname(viper.__file__),
    "data", "gauss_120.lut"))


class ViperTest(TestCase):
    """Test the Viper class.
    """
    def setUp(self):
        TestCase.setUp(self)

    def test_construct(self):
        """Test construction."""
        for kwds in ({}, {"actors": [viper.ScalarField(_vertices1, _cells1,
            _scalars1)]}, {"actors": viper.Mesh(_vertices1, _cells1)}):
            self.__test_construct(kwds)

    def __test_construct(self, kwds):
        vip = self.__construct(**kwds)
        self.assertNot(vip.outline_enabled)

        actors = kwds.get("actors")
        if actors:
            # Check that actors were added
            if not isinstance(actors, (list, tuple)):
                # Support supplying a single actor
                actors = [actors]
            self.assertEqual(vip.actors, actors)
            self.assertSortedEqual(vip.get_vtk_actors(), [a.vtk_actor for a in
                    actors])

    def test_clear(self):
        """Test clear method."""
        actors = [viper.Polygon([(0,0,0), (1,1,0)]), viper.Polygon([(1,1,0),
            (1,0,0)],), viper.Sphere((0.5,0.5,0.0), 0.2, 30, 20)]
        vip = self.__construct(actors=actors)
        ren_mock = Mock()
        self._set_attr(vip, "renderer", ren_mock)
        vip.clear()
        self.assertSeqEqual(vip.actors, [])
        for i, actor in enumerate(actors):
            ren_mock.mockCheckNamedCall(self, "RemoveActor", i, actor.vtk_actor)
        self.assertNot(vip.get_vtk_actors())


    def test_remove_actor(self):
        actor = viper.ScalarField(_vertices1, _cells1,
            _scalars1, actor_kwds={"lut": _lut_fname})
        vip = self.__construct(actors=[actor])
        ren_mock = Mock()
        self._set_attr(vip, "renderer", ren_mock)
        vip.remove_actor(actor)
        self.assertNotIn(actor, vip.actors)
        ren_mock.mockCheckNamedCall(self, "RemoveActor", 0, actor.vtk_actor)

    def test_remove_actor_with_scalarbar(self):
        """Test removing actor associated with scalarbar."""
        actor = viper.ScalarField(_vertices1, _cells1, _scalars1)
        vip = self.__construct(actors=[actor])
        ren_mock = Mock()
        self._set_attr(vip, "renderer", ren_mock)
        vip.get_scalarbar(vip.POS_N).should_show = True
        vip.get_scalarbar(vip.POS_N).actor = actor
        vip.remove_actor(actor)
        self.assertNotIn(actor, vip.actors)
        self.assertIs(vip.get_scalarbar(vip.POS_N).actor, None)
        _verify_scalarbar_visibility(self, vip.get_scalarbar(vip.POS_N), False)


    def test_reset_camera(self):
        """Test reset_camera method."""
        vip = self.__construct()

        renderer_mock = Mock()
        self._set_attr(vip, "renderer", renderer_mock)
        vip.reset_camera()
        self.assertEqual([c.name for c in renderer_mock.mockGetAllCalls()],
                ["ResetCamera", "ResetCameraClippingRange"])

        vip.reset_camera(clipping_range=False)
        self.assertEqual([c.name for c in renderer_mock.mockGetAllCalls()],
                ["ResetCamera", "ResetCameraClippingRange", "ResetCamera"])

        vip.reset_camera(center=False)
        self.assertEqual([c.name for c in renderer_mock.mockGetAllCalls()],
                ["ResetCamera", "ResetCameraClippingRange", "ResetCamera",
                    "ResetCameraClippingRange"])

    def test_azimuth(self):
        """Test azimuth method."""
        self.__check_camera_call("azimuth")

    def test_elevate(self):
        """Test elevate method."""
        self.__check_camera_call("elevate", "Elevation")

    def test_dolly(self):
        self.__check_camera_call("dolly")

    def test_set_viewangle(self):
        self.__check_camera_call("set_viewangle", "SetViewAngle")

    def __check_camera_call(self, operation, vtk_operation=None):
        if vtk_operation is None:
            vtk_operation = operation.capitalize()
        vip = self.__construct()
        cam_mock = Mock()
        self._set_attr(vip, "renderer", Mock(returnValues={"GetActiveCamera": cam_mock}))
        val = 2
        getattr(vip, operation)(val)
        cam_mock.mockCheckNamedCall(self, vtk_operation, 0, val)

    def test_camera_state(self):
        """Test save_camera and load_camera methods."""
        vip = self.__construct()

        cam = vip.renderer.GetActiveCamera()
        vip.azimuth(1)
        vip.elevate(2)
        vip.dolly(3)
        vip.set_viewangle(4)
        cam.SetClippingRange(10, 20)
        cam.SetDistance(10)
        cam.SetParallelProjection(not cam.GetParallelProjection())
        cam.SetViewAngle(10)
        cam.SetWindowCenter(2, 3)
        fname = self._get_tempfname()
        vip.save_camera(fname)
        # According to VTK documentation, these parameters are what matter
        (pos1, foc1, view1, clip1, dist1, par1, viewang1, center1) = (
                cam.GetPosition(), cam.GetFocalPoint(), cam.GetViewUp(),
                cam.GetClippingRange(), cam.GetDistance(),
                cam.GetParallelProjection(), cam.GetViewAngle(),
                cam.GetWindowCenter())
        vip.renderer.SetActiveCamera(vtk.vtkCamera())

        vip.load_camera(fname)
        cam = vip.renderer.GetActiveCamera()
        (pos2, foc2, view2, clip2, dist2, par2, viewang2, center2) = (
                cam.GetPosition(),
                cam.GetFocalPoint(), cam.GetViewUp(), cam.GetClippingRange(),
                cam.GetDistance(), cam.GetParallelProjection(),
                cam.GetViewAngle(), cam.GetWindowCenter())

        def assert_almost(val1, val2):
            if not isinstance(val1, (list, tuple)):
                val1 = [val1]
                val2 = [val2]
            for i, x in enumerate(val1):
                self.assertAlmostEqual(val2[i], x)

        assert_almost(pos2, pos1)
        assert_almost(foc2, foc1)
        assert_almost(view2, view1)
        assert_almost(clip1, clip2)
        assert_almost(dist1, dist2)
        self.assertEqual(par1, par2)
        self.assertEqual(viewang1, viewang2)
        self.assertEqual(center1, center2)


    def test_add_actor_labelable(self):
        """Test adding labelable actor."""
        actor = viper.Mesh(_vertices1, _cells1)
        actor.is_point_labels_enabled = True
        vip = self.__construct()
        vip.add_actor(actor)
        self.__verify_connected(vip, actor, True)
        lbl_mappers = [a for a in vip.get_vtk_actors2d() if a.GetMapper() is
                not None and a.GetMapper().IsA("vtkLabeledDataMapper")]
        self.assertEqual(len(lbl_mappers), 1, "Point labels not added")


    def test_remove_actor(self):
        """Test removing an actor."""
        actor = viper.Mesh(_vertices1, _cells1)
        vip = self.__construct([actor])
        vip.remove_actor(actor)
        self.assertNotIn(actor, vip.actors)
        self.assertNot(vip.get_vtk_actors())
        self.__verify_connected(vip, actor, False)


    def test_remove_actor_labelable(self):
        """Test removing labelable actor."""
        actor = viper.Mesh(_vertices1, _cells1)
        actor.is_point_labels_enabled = True
        vip = self.__construct([actor])
        vip.remove_actor(vip.actors[0])
        lbl_mappers = [a for a in vip.get_vtk_actors2d() if a.GetMapper() is not
                None and a.GetMapper().IsA("vtkLabeledDataMapper")]
        self.assertNot(lbl_mappers, "Point labels not removed")


    def __verify_connected(self, vip, actor, connected):
        for name in ("point_labels_enabled", "cell_labels_enabled"):
            name = "sig_%s" % name
            sig = getattr(actor, name)
            if connected:
                self.assert_(sig.is_obj_connected(vip), "Not connected to %s" % name)
            else:
                self.assertNot(sig.is_obj_connected(vip), "Still connected to %s" % name)


    def test_toggle_outline(self):
        """Test toggle_outline method."""
        vip = self.__construct(actors=[viper.Mesh(_vertices1, _cells1)])

        vip.toggle_outline()
        self.assert_(vip.outline_enabled)
        filters = [a for a in viper._util.get_vtk_actors(
            vip.renderer.GetActors()) if a.GetMapper().GetInputConnection(0,
                0).GetProducer().IsA("vtkOutlineFilter")]
        self.assertEqual(len(filters), 1)

        vip.toggle_outline()
        self.assertNot(vip.outline_enabled)
        filters = [a for a in viper._util.get_vtk_actors(
            vip.renderer.GetActors()) if a.GetMapper().GetInputConnection(0,
                0).GetProducer().IsA("vtkOutlineFilter")]
        self.assertEqual(len(filters), 0)

    def test_toggle_outline_add_actor(self):
        """Verify that new actors have outline after toggling."""
        vip = self.__construct()
        vip.toggle_outline()
        mesh = viper.Mesh(_vertices1, _cells1)
        vip.add_actor(mesh)
        filters = [a for a in viper._util.get_vtk_actors(
            vip.renderer.GetActors()) if a.GetMapper().GetInputConnection(0,
                0).GetProducer().IsA("vtkOutlineFilter")]
        self.assertEqual(len(filters), 1)


    def test_toggle_outline_remove_actor(self):
        """Verify that outlines are removed with actors."""
        mesh = viper.Mesh(_vertices1, _cells1)
        vip = self.__construct()
        vip.toggle_outline()
        vip.add_actor(mesh)
        vip.remove_actor(mesh)

        self.assertNot(list(viper._util.get_vtk_actors(
            vip.renderer.GetActors())))


    def test_get_vtk_actors(self):
        """Test get_vtk_actors method."""
        vip = self.__construct()
        self.assertSeqEqual(vip.get_vtk_actors(), [])
        actor = viper.Polygon([(0, 0, 0)])
        vip.add_actor(actor)
        self.assertSeqEqual(vip.get_vtk_actors(), [actor.vtk_actor])

    def test_get_vtk_actors_none(self):
        """Verify that get_vtk_actors handles None being returned from
        GetActors."""
        vip = self.__construct()
        self._set_attr(vip.renderer, "GetActors", mock.FunctionMock(None))
        self.assertSeqEqual(vip.get_vtk_actors(), [])


    def test_get_scalarbar(self):
        """Test get_scalarbar method."""
        vip = self.__construct([viper.Mesh(_vertices1, _cells1)])
        for pos in (vip.POS_N, vip.POS_S, vip.POS_E, vip.POS_W):
            bar = vip.get_scalarbar(pos)
            self.assert_(isinstance(bar, viper._viper._Scalarbar))

    def test_get_scalarbar_invalid_pos(self):
        """Test get_scalarbar method providing an invalid position."""
        vip = self.__construct([viper.Mesh(_vertices1, _cells1)])
        self.assertRaises(ValueError, vip.get_scalarbar, 5)


    def __construct(self, *args, **kwds):
        def modified():
            self.__modified = True
        vip = viper.Viper(*args, **kwds)
        self.__modified = False
        vip.sig_modified.connect(modified)
        return vip


def _verify_scalarbar_visibility(test, bar, visible):
    test.assertEqual(bar.is_visible, visible)
    vtk_is_visible = test._get_privattr(bar, "bar_actor").GetVisibility()
    if visible:
        test.assert_(vtk_is_visible)
    else:
        test.assertNot(vtk_is_visible)


class ScalarbarTest(TestCase):
    """Test _Scalarbar."""
    def test_construct(self):
        """Test constructor."""
        bar = self.__construct()
        self.assertIs(bar.actor, None)
        _verify_scalarbar_visibility(self, bar, False)


    def test_set_actor(self):
        """Test setting actor."""
        actor = viper.ScalarField(_vertices1, _cells1, _scalars1)
        bar = self.__construct(actor=actor)
        bar.actor = actor
        self.assertIs(bar.actor, actor)
        _verify_scalarbar_visibility(self, bar, False)
        self.assert_(actor.vtk_actor.GetMapper().HasObserver("ModifiedEvent"))


    def test_visibility(self):
        """Test visibility state."""
        bar = self.__construct()
        actor = viper.ScalarField(_vertices1, _cells1, _scalars1)
        bar.actor = actor
        _verify_scalarbar_visibility(self, bar, False)

        bar.should_show = True
        _verify_scalarbar_visibility(self, bar, True)

        bar.actor = None
        bar.actor = actor
        _verify_scalarbar_visibility(self, bar, True)


    def test_unset_actor(self):
        """Test unsetting actor."""
        actor = viper.ScalarField(_vertices1, _cells1, _scalars1)
        bar = self.__construct(actor=actor)
        bar.should_show = True
        bar.actor = actor
        _verify_scalarbar_visibility(self, bar, True)
        bar.actor = None
        _verify_scalarbar_visibility(self, bar, False)
        self.assertIs(bar.actor, None)
        self.assertNot(actor.vtk_actor.GetMapper().HasObserver("ModifiedEvent"))


    def test_change_lut(self):
        """Change an actor's LUT."""
        actor = viper.Mesh(_vertices1, _cells1)
        bar = self.__construct(actor=actor)
        # After setting the associated actors' LUT, the scalarbar should update
        # itself
        actor.lut = viper.LookupTable()
        self.assertEqual(bar.lut, actor.lut)
        self.assertEqual(bar.lut.range, actor.scalar_range)


    def __construct(self, pos=viper.Viper.POS_N, actor=None):
        vip = self.__vip = viper.Viper()
        bar = viper._viper._Scalarbar(vip, pos)
        bar.actor = actor
        return bar



def _verify_lut(test, actor_lut, lut_src):
    """Verify that a LUT corresponds to a source specification.
    @param actor_lut: The actor's L{LUT<viper.LookupTable>}.
    @param lut_src: The source specification to compare against. This is a
    color table or a filename.
    """
    if lut_src is None:
        test.assertIs(actor_lut, viper.get_default_lut())
        return

    if isinstance(lut_src, basestring):
        # Filename
        lut_file = file(lut_src)
        it = iter(lut_file)
        # Skip the first line
        it.next()
        lut_src = []
        for l in it:
            row = [float(x) for x in l.strip().split()]
            if len(row) == 4:
                lut_src.append(row)
    test.assertEqual(actor_lut.vtk_lut.GetNumberOfColors(),
            len(lut_src))
    for i, val in enumerate(lut_src):
        vtk_val = actor_lut.vtk_lut.GetTableValue(i)
        test.assertIsNot(vtk_val, None, "No lookup value for %d" % i)
        for i in range(4):
            test.assertAlmostEqual(vtk_val[i], val[i], places=2)


class _ActorTestBase(TestCase):
    """Baseclass for testing actors."""
    def test_construct_actor_kwds(self):
        for kwds in ({}, {"opacity": 0.5, "color": (0, 0, 0), "lut":
            viper.LookupTable(_lut_fname)},):
            actor = self._construct(actor_kwds=kwds)
            self.assert_(actor.is_visible)
            if "opacity" in kwds:
                self.__verify_opacity(actor, kwds["opacity"])
            if "color" in kwds:
                self.__verify_color(actor, kwds["color"])
            else:
                self.__verify_color(actor, (0, 0, 0))
            lut = kwds.get("lut")
            if lut is None or lut is viper._viper._STOCK_LUT:
                self.assertIs(actor.lut, viper.get_default_lut())
            else:
                # Match against the gauss 120 LUT file
                self.assertEqual(actor.lut.vtk_lut.GetNumberOfColors(),
                        256)


    def test_get_property(self):
        """Test the get_property method."""
        actor = self._construct()
        prop_mock = mock.Mock(returnValues={"GetSomeProperty": 1})
        self._set_attr(actor.vtk_actor, "GetProperty",
                mock.FunctionMock(returnValue=prop_mock))
        self.assertEqual(actor.get_property("SomeProperty"), 1)

    def test_get_property_bad_name(self):
        """Test invoking get_property with bad property name."""
        actor = self._construct()
        prop_mock = mock.Mock(dontMock=["GetSomeProperty"])
        self._set_attr(actor.vtk_actor, "GetProperty",
                mock.FunctionMock(returnValue=prop_mock))
        self.assertRaises(ValueError, actor.get_property, "SomeProperty")


    def test_set_property(self):
        """Test the set_property method."""
        actor = self._construct()
        prop_mock = mock.Mock()
        self._set_attr(actor.vtk_actor, "GetProperty",
                mock.FunctionMock(returnValue=prop_mock))

        for args in ((), (1,), (1, 2)):
            actor.set_property("SomeProperty", *args)
            prop_mock.mockCheckNamedCall(self, "SetSomeProperty", -1, *args)

    def test_set_property_bad_name(self):
        """Test invoking set_property with bad property name."""
        actor = self._construct()
        prop_mock = mock.Mock(dontMock=["SetSomeProperty"])
        self._set_attr(actor.vtk_actor, "GetProperty",
                mock.FunctionMock(returnValue=prop_mock))
        self.assertRaises(ValueError, actor.set_property, "SomeProperty",
                "val")


    def test_update_source(self):
        """Test update_source method."""
        actor = self._construct()
        if actor.source is None:
            return
        mtime_prev = actor.source.GetMTime()
        actor.update_source()
        self.assertNotEqual(actor.source.GetMTime(), mtime_prev)

    def test_update_source_new_data(self):
        """Test update_source method supplying new data."""
        actor = self._construct()
        self._test_update_source_new_data(actor)

    def _test_update_source_new_data(self, actor):
        """Default implementation."""


    def test_opacity(self):
        """Test the opacity property."""
        actor = self._construct()
        actor.opacity = 0.5
        self.__verify_opacity(actor, 0.5)

    def test_color(self):
        """Test the color property."""
        actor = self._construct()
        actor.color = (1, 1, 1)
        self.__verify_color(actor, (1, 1, 1))


    def test_is_visible(self):
        """Test the is_visible property."""
        actor = self._construct()
        actor.is_visible = False
        self.assertNot(actor.is_visible)
        self.assertNot(actor.vtk_actor.GetVisibility())
        actor.is_visible = True
        self.assert_(actor.is_visible)
        self.assert_(actor.vtk_actor.GetVisibility())


    def test_set_lut(self):
        """Test setting lut."""
        actor = self._construct()
        for src in (None, _lut_fname, viper._viper._STOCK_LUT):
            if src is not None:
                lut = viper.LookupTable(src)
            else:
                lut = None
            actor.lut = lut
            _verify_lut(self, actor.lut, src)
            self.assertIs(actor.vtk_actor.GetMapper().GetLookupTable(),
                    actor.lut.vtk_lut)


    def __verify_opacity(self, actor, opacity):
        self.assertEqual(actor.opacity, opacity)
        self.assertEqual(actor.vtk_actor.GetProperty().GetOpacity(), opacity)

    def __verify_color(self, actor, color):
        self.assertEqual(actor.color, color)
        self.assertEqual(actor.vtk_actor.GetProperty().GetColor(), color)


def _verify_input_conn(test, obj, input_classname):
    """Verify that a VTK object has input from a certain class."""
    test.assert_(obj.GetInputConnection(0, 0).GetProducer().IsA(input_classname))


class _LabelableTestBase(object):
    """Mixin for testing ILableables."""
    def test_labelable_construct(self):
        """Test constructor."""
        actor = self._construct()
        self.assert_(isinstance(actor, viper.ILabelable),
                "%s does not implement ILabelable" % actor.__class__.__name__)
        self.assertNot(actor.is_point_labels_enabled)
        self.assertNot(actor.is_cell_labels_enabled)


class _MeshTestBase(_ActorTestBase, _LabelableTestBase):
    """Baseclass for _MeshBase tests."""
    def test_mesh_construct_cell_format(self):
        """Test that cell data format is verified."""
        idtype = viper.get_vtk_idtype()
        if idtype == "int32":
            dtype = "int64"
        else:
            dtype = "int32"
        self.assertRaises(ValueError, self._construct, vertices=_vertices1,
                cells=numpy.array(_cells1, dtype=dtype))


class ScalarFieldTest(_MeshTestBase):
    """Test the ScalarField class.
    """
    def test_construct(self):
        """Test construction."""
        for kwds in ({}, {"scalar_range": (0, 0)}, {"vertices": _vertices3,
            "scalars": _scalars3, "cells": _cells3, "cutplane_origin": (0, 0,
                0.1)},):
            self.__test_construct(kwds)

    def test_construct_displacement(self):
        """Test constructing with displacement."""
        displacement = 0.2*_vertices1.copy()
        self.__test_construct({"displacement": displacement})

    def test_construct_displacement_and_cutplane(self):
        """Test constructing with displacement and cutplane."""
        displacement = 0.2*_vertices3.copy()
        self.__test_construct({"vertices": _vertices3, "scalars": _scalars3,
            "cells": _cells3, "cutplane_origin": (0, 0, 0.1), "displacement":
            displacement})

    def __test_construct(self, kwds):
        field = self._construct(**kwds)
        actor_kwds = kwds.get("actor_kwds", {})
        self.assertNot(field._mapper.GetUseLookupTableScalarRange())
        scalar_range = kwds.get("scalar_range")
        if scalar_range is None:
            # Unspecified scalar range means to deduce from scalars
            scalar_range = field.scalars.min(), field.scalars.max()
        self.assertEqual(field.scalar_range, scalar_range)
        cut_origin = kwds.get("cutplane_origin")
        displacement = kwds.get("displacement")
        if displacement is None:
            if cut_origin is not None:
                _verify_input_conn(self, field._pipeline_start, "vtkCutter")
        else:
            if cut_origin is None:
                _verify_input_conn(self, field._pipeline_start, "vtkWarpVector")
            else:
                cut_filter = field._pipeline_start.GetInputConnection(0,
                        0).GetProducer()
                _verify_input_conn(self, cut_filter, "vtkWarpVector")
                _verify_input_conn(self, field._pipeline_start, "vtkCutter")


    def test_update_displacement(self):
        """Test update_displacement method."""
        displacement = 0.2*_vertices1.copy()
        field = self._construct(displacement = displacement)
        disp_source = field._grid.GetPointData().GetVectors()
        mtime_prev = disp_source.GetMTime()
        field.update_displacement()
        self.assertNotEqual(disp_source.GetMTime(), mtime_prev)


    def test_set_scalar_range(self):
        """Test setting scalar range."""
        field = self._construct()
        range = (10, 100)
        field.scalar_range = range
        self.assertEqual(field.scalar_range, range)
        self.assertEqual(field.vtk_actor.GetMapper().GetScalarRange(), range)


    def _test_update_source_new_data(self, actor):
        """Called by baseclass to test update_source method with new data."""
        scalars = numpy.linspace(1, 0, len(_vertices1))
        actor.update_source(scalars)
        self.assertIs(actor.scalars, scalars)
        self.assertSeqEqual(list(viper._util.iterate_vtk_array(actor.source)),
                actor.scalars)


    def _construct(self, vertices=_vertices1, cells=_cells1, scalars=_scalars1,
            scalar_range=None, cutplane_origin=None, displacement=None,
            actor_args=(), actor_kwds={}):
        return viper.ScalarField(vertices, cells, scalars, scalar_range=
                scalar_range, cutplane_origin=cutplane_origin, displacement=
                displacement, actor_args=actor_args, actor_kwds=actor_kwds)


class VectorFieldTest(_ActorTestBase):
    """Test the VectorField class."""
    def test_construct(self):
        """Test construction."""
        for kwds in ({}, {"cells": _cells2, "actor_kwds": {"lut":
            viper.LookupTable()}}):
            self.__test_construct(kwds)

    def __test_construct(self, kwds):
        vector = self._construct(**kwds)
        # By default, _scalars2 is used for directions parameter
        rmin, rmax = self._get_privattr(vector, "comp_range")(_scalars2)
        self.assertEqual(vector._mapper.GetScalarRange(), (rmin, rmax))
        self.assertNot(vector._mapper.GetUseLookupTableScalarRange())


    def _test_update_source_new_data(self, actor):
        """Called by baseclass to test update_source method with new data."""
        dirs = _scalars2
        actor.update_source(dirs)
        self.assertIs(actor.directions, dirs)
        self.assertSeqEqual(list(viper._util.iterate_vtk_array(actor.source)),
                numpy.ravel(actor.directions))


    def _construct(self, vertices=_vertices2, directions=_scalars2, cells=None,
            actor_args=(), actor_kwds={}):
        return viper.VectorField(vertices, directions, cells=cells,
                actor_args=actor_args, actor_kwds=actor_kwds)


class SphereTest(_ActorTestBase, _LabelableTestBase):
    """Test the Sphere class."""
    def test_construct(self):
        """Test construction."""
        for kwds in ({}, {"theta_res": 10}, {"phi_res": 10}):
            self.__test_construct(kwds)

    def __test_construct(self, kwds):
        sphere = self._construct(**kwds)
        point, radius, theta_res, phi_res = (kwds.get("point"),
                kwds.get("radius"), kwds.get("theta_res"), kwds.get("phi_res"))
        src = sphere.source
        if point is None:
            point = (0, 0, 0)
        if radius is None:
            radius = 0.2

        self.assertEqual(src.GetCenter(), point)
        self.assertEqual(src.GetRadius(), radius)
        if theta_res is not None:
            self.assertEqual(sphere.source.GetThetaResolution(), theta_res)
        if phi_res is not None:
            self.assertEqual(sphere.source.GetPhiResolution(), phi_res)
        self.assertIs(sphere.vtk_actor.GetMapper().GetInputConnection(0,
            0).GetProducer(), sphere.source)


    def _construct(self, point=None, radius=None, theta_res=None, phi_res=None,
            actor_args=(), actor_kwds={}):
        if point is None:
            point = (0, 0, 0)
        if radius is None:
            radius = 0.2
        kwds = {"actor_args": actor_args, "actor_kwds": actor_kwds}
        if theta_res is not None:
            kwds["theta_res"] = theta_res
        if phi_res is not None:
            kwds["phi_res"] = phi_res

        return viper.Sphere(point, radius, **kwds)


class PolygonTest(_ActorTestBase):
    """Test the Polygon class."""
    def test_construct(self):
        """Test construction."""
        for kwds in ({},):
            self.__test_construct(kwds)

    def __test_construct(self, kwds):
        poly = self._construct(**kwds)


    def _construct(self, vertices=None, actor_args=(), actor_kwds={}):
        if vertices is None:
            vertices = [(0, 0, 0),]
        kwds = {"actor_args": actor_args, "actor_kwds": actor_kwds}

        return viper.Polygon(vertices, **kwds)



class PngWriterTest(TestCase):
    """Test the PngWriter class."""
    def test_capture_frame(self):
        """Test capture_frame method."""
        ren_mock = mock.Mock()
        render_large_out = mock.Mock()
        render_large_factory = mock.MockFactory(mockKwds={"returnValues":
            {"GetOutputPort": render_large_out}})
        self._set_attr(vtk, "vtkRenderLargeImage", render_large_factory)
        png_factory = mock.MockFactory()
        self._set_attr(vtk, "vtkPNGWriter", png_factory)
        vip = viper.Viper()
        self._set_attr(vip, "renderer", ren_mock)
        writer = viper.PngWriter(vip)
        
        render_large = render_large_factory.mock_constructed[0]
        png = png_factory.mock_constructed[0]
        render_large.mockCheckCall(self, 0, "SetInput", vip.renderer)
        render_large.mockCheckCall(self, 1, "SetMagnification", 1)
        png.mockCheckCall(self, 0, "SetInputConnection", render_large_out)

        fname = self._get_tempfname()
        writer.capture_frame(fname)

        ren_mock.mockCheckCall(self, 0, "Render")
        render_large.mockCheckCall(self, 3, "Update")
        render_large.mockCheckCall(self, 4, "Modified")
        png.mockCheckCall(self, 1, "SetFileName", fname)
        png.mockCheckCall(self, 2, "Write")


class MovieMakerTest(TestCase):
    """Test the MovieMaker class."""
    def test_construct(self):
        fname = self._get_tempfname()
        maker = self.__construct(fname, fps=35)
        self.assertEqual(maker.filename, fname)
        self.assertEqual(maker.fps, 35)

    def test_construct_no_encoder(self):
        """Test constructing with no suitable encoder installed."""
        fname = self._get_tempfname()
        self._set_attr(viper._util, "resolve_path", mock.FunctionMock(raises=viper._util.NotFound))
        self.assertRaises(viper.ViperError, self.__construct, fname)


    def test_capture_frame(self):
        maker = self.__construct()
        maker.capture_frame()
        self.assertEqual(len(self.__png_writer.mockGetNamedCalls("capture_frame")),
                1)


    def test_make_movie(self):
        check_call = mock.FunctionMock(0)
        self._set_attr(subprocess, "check_call", check_call)
        maker = self.__construct()
        self.assertRaises(viper.NoFramesError, maker.make_movie)

        maker.capture_frame()
        maker.make_movie()
        calls = check_call.mockGetNamedCalls("__call__")
        self.assertEqual(len(calls), 1)
        self.assertEqual(calls[0].args[0][0], "mencoder")

    def test_make_movie_encoder_fail(self):
        check_call = mock.FunctionMock(raises=subprocess.CalledProcessError(-1, "cmd"))
        self._set_attr(subprocess, "check_call", check_call)
        maker = self.__construct()
        maker.capture_frame()
        self.assertRaises(viper.ViperError, maker.make_movie)


    def __construct(self, fname=None, **kwds):
        if fname is None:
            fname = self._get_tempfname()
        png_factory = mock.MockFactory()
        self._set_attr(viper._viper, "PngWriter", png_factory)
        maker = viper.MovieMaker(mock.Mock(), fname, **kwds)
        self.__png_writer = png_factory.mock_constructed[0]
        return maker


class MiscTest(TestCase):
    """Test miscellaneous."""
    def test_set_default_lut(self):
        """Test function set_default_lut."""
        # Make sure the default LUT is set at first
        for i, val in enumerate(viper._viper._STOCK_LUT):
            lut = viper.get_default_lut()
            self.assertSeqEqual(lut.vtk_lut.GetTableValue(i), val)
        viper.set_default_lut(None)
        lut = viper.get_default_lut()
        for i, val in enumerate(viper._viper._STOCK_LUT):
            self.assertSeqEqual(lut.vtk_lut.GetTableValue(i), val)

        viper.set_default_lut(_lut_fname)
        self.assertEqual(viper.get_default_lut().vtk_lut.GetNumberOfColors(),
                256)


class _WidgetTestBase(TestCase):
    """Metaclass for testing subclasses of WidgetBase."""
    def setUp(self):
        TestCase.setUp(self)
        self._set_attr(vtk, "vtkRenderWindow", mock.MockFactory())
        self._set_attr(vtk, "vtkRenderWindowInteractor", mock.MockFactory())


    def test_construct(self):
        """Test construction."""
        win = self._construct()
        self.assertNot(win.is_offscreen)


    def test_enable_offscreen(self):
        """Test enabling offscreen rendering."""
        self._set_attr(vtk, "vtkGraphicsFactory", mock.MockFactory())
        self._set_attr(vtk, "vtkImagingFactory", mock.MockFactory())
        win = self._construct()
        win.is_offscreen = True

        if _on_windows:
            win.vtk_window.mockCheckNamedCall(self, "SetOffScreenRendering", 0,
                    True)
        else:
            grfx_fact = vtk.vtkGraphicsFactory.mock_constructed[0]
            imging_fact = vtk.vtkImagingFactory.mock_constructed[0]
            grfx_fact.mockCheckNamedCall(self, "SetOffScreenOnlyMode", 0, True)
            grfx_fact.mockCheckNamedCall(self, "SetUseMesaClasses", 0, True)
            imging_fact.mockCheckNamedCall(self, "SetUseMesaClasses", 0, True)
    
    def test_disable_offscreen(self):
        """Test disabling offscreen rendering."""
        self._set_attr(vtk, "vtkGraphicsFactory", mock.MockFactory())
        self._set_attr(vtk, "vtkImagingFactory", mock.MockFactory())
        win = self._construct()
        win.is_offscreen = True
        win.is_offscreen = False
        if _on_windows:
            win.vtk_window.mockCheckNamedCall(self, "SetOffScreenRendering", 0,
                    True)
        else:
            # The factories are constructed for each invocation
            grfx_fact = vtk.vtkGraphicsFactory.mock_constructed[-1]
            imging_fact = vtk.vtkImagingFactory.mock_constructed[-1]
            grfx_fact.mockCheckNamedCall(self, "SetOffScreenOnlyMode", -1, False)
            grfx_fact.mockCheckNamedCall(self, "SetUseMesaClasses", -1, False)
            imging_fact.mockCheckNamedCall(self, "SetUseMesaClasses", -1, False)


class VtkWindowTest(_WidgetTestBase):
    """Test VtkWindow."""
    def _construct(self):
        vip = viper.Viper()
        return viper.VtkWindow(vip)


class CellZerosTest(TestCase):
    """Test CellZeros factory function."""
    def test_construct(self):
        self.assertEqual(viper.CellZeros((1, 1, 1)), numpy.zeros((1, 1, 1),
            dtype=viper._util.get_vtk_idtype()))


class LookupTableTest(TestCase):
    """Test LookupTable class."""
    def test_construct(self):
        lut = viper.LookupTable()
        self.assertEqual(lut.vtk_lut.GetRange(), (0, 1))
        lut = viper.LookupTable(range=(0.5, 1))
        self.assertEqual(lut.vtk_lut.GetRange(), (0.5, 1))
